import { createButton, createContainer } from '@/common/helpers'
import type { GameSummary, MatchSessionDto } from '@/common/interfaces'
import { EventEmitter, type Application, type Container } from 'pixi.js'
import { createText, whiteStyle, yellowStyle } from './utilities/fonts'

export class GameSummayView extends EventEmitter {
  public width: number
  public height: number
  container!: Container
  layer!: Container
  gameSummary!: GameSummary
  matchState!: MatchSessionDto
  type: 'round' | 'match' = 'round'

  constructor(app: Application) {
    super()
    this.width = 500
    this.height = 400
    this.container = createContainer({
      width: this.width,
      height: this.height,
      x: app.canvas.width / 2 - this.width / 2,
      y: app.canvas.height / 2 - this.height / 2,
      parent: app.stage,
      alpha: 0.7,
      color: 0x121212,
    })
    this.layer = createContainer({
      width: this.width,
      height: this.height,
      parent: this.container,
    })
    this.container.visible = false
  }

  setGameSummary(data: any, type: 'round' | 'match') {
    this.type = type
    this.matchState = data.sessionState
    this.gameSummary = data.lastGame
    this.render()
    this.container.visible = true
  }

  renderTitle(y: number = 20, title: string): number {
    const text = createText({
      text: title,
      x: this.width / 2,
      y,
      style: yellowStyle(24),
    })
    this.layer.addChild(text)
    return y + 24
  }

  renderWinner(y: number): number {
    let line = y + 12
    this.layer.addChild(
      createText({
        text: `Winner: ${this.gameSummary.winner.name}`,
        x: this.width / 2,
        y: line,
        style: whiteStyle(20),
      }),
    )

    if (this.gameSummary.isBlocked) {
      line += 30
      this.layer.addChild(
        createText({
          text: '(Blocked)',
          x: this.width / 2,
          y: line,
          style: whiteStyle(),
        }),
      )
    }

    line += 30
    this.layer.addChild(
      createText({
        text: `Points this round: ${this.gameSummary.winner.score}`,
        x: this.width / 2,
        y: line,
        style: whiteStyle(20),
      }),
    )
    return line + 16
  }

  renderScores(y: number): number {
    const scores = this.matchState.scoreboard
    // this.type === 'round'
    //   ? this.gameSummary.score
    //   : this.matchState.scoreboard.map((d) => ({ name: d[0], score: d[1] }))
    let line = y + 30
    scores.forEach((score: any) => {
      line = line + 30
      this.layer.addChild(
        createText({
          text: `${score.name}:`,
          x: 130,
          y: line,
          style: whiteStyle(18),
          align: 'left',
        }),
      )
      this.layer.addChild(
        createText({
          text: `${score.score}`,
          x: 330,
          y: line,
          style: whiteStyle(18),
          align: 'right',
        }),
      )
    })
    return line
  }

  renderButtons() {
    if (this.type === 'round') {
      this.layer.addChild(
        createButton({
          text: 'Next',
          dimension: {
            x: this.width / 2 - 25,
            y: this.height - 50,
            width: 60,
            height: 25,
          },
          action: () => {
            this.emit('nextClick', { sessionState: this.matchState })
            this.container.visible = false
          },
          parent: this.layer,
        }),
      )
    } else {
      this.layer.addChild(
        createButton({
          text: 'Finish',
          dimension: {
            x: this.width / 2 - 25,
            y: this.height - 50,
            width: 60,
            height: 25,
          },
          action: () => {
            this.emit('finishClick', this.gameSummary)
            this.container.visible = false
          },
          parent: this.layer,
        }),
      )
    }
  }

  render() {
    const title: string = this.type === 'round' ? 'Round Summary' : 'Match Finished!'
    this.layer.removeChildren()
    let y = this.renderTitle(30, title.toUpperCase())
    y = this.renderWinner(y)
    this.renderScores(y)
    this.renderButtons()
  }
}
